(declare (standard-bindings)
	 (extended-bindings)
	 (not safe)
	 (block))

(declare (not core))
(include "points.scm")
(declare (core))

(define (show-boundary-coefficients v)
  (with-access
   v (Linear-finite-element-vector data space)
   (with-access
    space (Linear-finite-element-space triangulation)
    (let ((data data)
	  (triangulation triangulation))
      (Triangulation-for-each-boundary-vertex
       (lambda (v)
	 (with-access
	  v (Vertex index)
	  (display (f64vector-ref data index)) (newline)))
       triangulation)))))

(define (show-class object)
  (Class-name (object->class object)))

(define number-of-refinements 7)

(define triangulations (Triangulation->refined-triangulations coarse-square-triangulation number-of-refinements))
(define spaces         (triangulations->linear-finite-element-spaces triangulations))
(define projectors     (linear-finite-element-spaces->projectors spaces))
(define injectors      (projectors->injectors projectors))

(define fes (vector-ref spaces number-of-refinements))

(define t (vector-ref triangulations number-of-refinements))

(Triangulation-consistent? t)

(define cg-data (CG-build-data t fes))

(define fine-t (refine t))

(define fine-fes (Triangulation->Linear-finite-element-space fine-t))

(define L2-inner-product-operator (->Linear-finite-element-operator fine-fes))

(add-mass-terms! L2-inner-product-operator (lambda (p) 1.0))

(set! L2-inner-product-operator (->Dirichlet-linear-finite-element-operator L2-inner-product-operator))

(define H1-inner-product-operator (->Linear-finite-element-operator fine-fes))

(add-stiffness-terms! H1-inner-product-operator (lambda (p) 1.0))

(add-mass-terms! H1-inner-product-operator (lambda (p) 1.0))

(set! H1-inner-product-operator (->Dirichlet-linear-finite-element-operator H1-inner-product-operator))

(define coarse->fine-operator (Operator-adjoint (coarse<-fine fes fine-fes)))

(define pi (* 4 (atan 1)))

(define symmetric-tests
  (list
   (instantiate Problem-data
		description: "a=1, c=0, u=sin(\\pi x)*sin(3\\pi y), Dirichlet"
		a: (lambda (p) 1.0)
		c: (lambda (p) 0.0)
		u: (lambda (p)
		     (let ((x (Point-x p))
			   (y (Point-y p)))
		       (fl* (flsin (fl* pi x)) (flsin (fl* 3. pi y)))))
		f: (lambda (p)
		     (let ((x (Point-x p))
			   (y (Point-y p)))
		       (fl+ (fl* pi pi (flsin (fl* pi x)) (flsin (fl* 3. pi y)))
			    (fl* (flsin (fl* pi x)) 3. 3. pi pi (flsin (fl* 3. pi y)))))))
   (instantiate Problem-data
		description: "a=1, c=0, u=x*(1-x)*y*(1-y), Dirichlet"
		a: (lambda (p) 1.0)
		c: (lambda (p) 0.0)
		u: (lambda (p)
		     (let ((x (Point-x p))
			   (y (Point-y p)))
		       (fl* x (fl- 1. x) y (fl- 1. y))))
		f: (lambda (p)
		     (let ((x (Point-x p))
			   (y (Point-y p)))
		       (fl+ (fl* 2. y (fl- 1. y))
			    (fl* 2. x (fl- 1. x))))))))

(define (run-test check solver-thunk)
  
  (define (format-for-display x)
    (if (fl< x 1.0)
	(let ((x (fl+ 1.0 x)))
	  (let ((string-x (number->string x)))
	    (string-append (substring string-x 1 (fxmin 12 (string-length string-x)))
			   (make-string (fx- 12 (fxmin 12 (string-length string-x))) #\0))))
	(number->string x)))
  
  (let* ((at-start (##process-statistics))
	 (result (solver-thunk))
	 (at-end (##process-statistics))
	 (user+sys-ns (fl+ (fl- (f64vector-ref at-end 0) (f64vector-ref at-start 0))
			   (fl- (f64vector-ref at-end 1) (f64vector-ref at-start 1))))
	 (gc-ns (fl+ (fl- (f64vector-ref at-end 3) (f64vector-ref at-start 3))
		     (fl- (f64vector-ref at-end 4) (f64vector-ref at-start 4))))
	 (difference (Vector-subtract (Operator-apply coarse->fine-operator result) check)))
    
    (for-each display (list #\tab
			    (format-for-display (sqrt (Vector-inner-product difference L2-inner-product-operator difference)))))
    (for-each display (list #\tab
			    (format-for-display (sqrt (Vector-inner-product difference H1-inner-product-operator difference)))))
    (for-each display (list #\tab
			    (round (fl* 1000. (fl- user+sys-ns gc-ns)))))
    (for-each display (list #\tab
			    (round (fl* 1000. gc-ns))
			    #\newline))))


(define (run-multigrid-tests tests)
  
  (for-each (lambda (problem-data)
	      (with-access
	       problem-data (Problem-data a c f gamma g u)

	       ;; First we create the Neumann operators and RHS.

	       (let* ((check            (->Dirichlet-linear-finite-element-vector! (interpolate-procedure->Linear-finite-element-vector u fine-fes)))
		      (operator         (build-operator a c gamma fes))
		      (rhs              (build-rhs f g fes))
		      (right-hand-sides (vector-map ->Dirichlet-linear-finite-element-vector! (most-refined-rhs->all-refined-rhs rhs projectors)))
		      (operators        (vector-map ->Dirichlet-linear-finite-element-operator (most-refined-operator->all-refined-operators operator projectors injectors))))
		 
		 (for-each display (list #\newline (Problem-data-description problem-data) #\newline))
		 
		 (for-each (lambda (smoother-maker smoother-name)
			     (let ((smoothers! (vector-map (lambda (smoother) (lambda (rhs z) (->Dirichlet-linear-finite-element-vector! (smoother rhs z)))) (make-smoothers operators smoother-maker))))
			       (for-each (lambda (p)
					   (for-each (lambda (m1)
						       
						       (define (iota m)
							 (let loop ((result '()) (m m))
							   (if (fx< m 0)
							       result
							       (loop (cons m result) (fx- m 1)))))
						       
						       (for-each (lambda (m2)
								   (for-each (lambda (r)
									       (let ((mg-data (instantiate Multigrid-data
													   triangulations:   triangulations
													   spaces:           spaces
													   projectors:       projectors
													   injectors:        injectors
													   operators:        operators
													   smoothers!:       smoothers!)))
										 
										 (MG-add-solver-information! mg-data smoother-name r m1 m2 p)
										 (display (Multigrid-data-solver-description mg-data))
										 (run-test check (lambda () (MG-solver mg-data right-hand-sides)))))
									     '(1 2 3 4)))
								 (list m1)));(iota m1)))
						     '(1 2 3 4)))
					 '(1 2))))
			   (list richardson-smoother!
				 jacobi-smoother!
				 gauss-seidel-smoother!
				 )
			   (list "  Richardson"
				 "      Jacobi"
				 "Gauss-Seidel"
				 )))))
	    tests))

(define (run-cg-tests tests)
  
  (for-each display (list "The number of refinements is " number-of-refinements #\newline))
  
  (for-each (lambda (problem-data)
	      
	      (for-each display (list #\newline (Problem-data-description problem-data) #\newline))
	      
	      (CG-add-Dirichlet-operator-and-rhs-information! cg-data problem-data)
	      
	      (for-each (lambda (preconditioner preconditioner-name)
			  
			  (define check (->Dirichlet-linear-finite-element-vector! (interpolate-procedure->Linear-finite-element-vector (Problem-data-u problem-data) fine-fes)))
			  
			  (CG-add-solver-information! cg-data
						      (preconditioner (Conjugate-gradient-data-operator cg-data))
						      preconditioner-name 
						      2000
						      (expt (Vector-dimension (Conjugate-gradient-data-right-hand-side cg-data)) -2.0))
			  (display (Conjugate-gradient-data-solver-description cg-data))
			  (run-test check (Conjugate-gradient-data-solver cg-data)))
			
			(list (lambda (m) identity-operator)
			      diagonal-preconditioner
			      gauss-seidel-preconditioner
			      )
			'("(Identity preconditioner)    "
			  "(Diagonal preconditioner)    "
			  "(Gauss-Seidel preconditioner)"
			  )))
	    tests))
